class PeaqValueError(ValueError):
    def __init__(self, message: str):
        self.message = message

class PeaqKeyError(KeyError):
    def __init__(self, message: str):
        self.message = message


NON_HOUR = "Charging stopped"
CAUTION_HOUR = "Charging-permittance degraded"
CHARGING_PERMITTED = "Charging permitted"
# moved to core

CAUTIONHOURTYPE_SUAVE = "Suave"
CAUTIONHOURTYPE_INTERMEDIATE = "Intermediate"
CAUTIONHOURTYPE_AGGRESSIVE = "Aggressive"

CAUTIONHOURTYPE = {
    CAUTIONHOURTYPE_SUAVE: 0.75,
    CAUTIONHOURTYPE_INTERMEDIATE: 0.5,
    CAUTIONHOURTYPE_AGGRESSIVE: 0.4
}

CAUTIONHOURTYPE_NAMES =[
    CAUTIONHOURTYPE_SUAVE,
    CAUTIONHOURTYPE_INTERMEDIATE,
    CAUTIONHOURTYPE_AGGRESSIVE
]


import logging
from ...util import try_parse
from .hubmember import HubMember
from .const import CARPOWERSENSOR

_LOGGER = logging.getLogger(__name__)


class CarPowerSensor(HubMember):
    def __init__(
            self,
            data_type: type,
            listenerentity=None,
            initval=None,
            powermeter_factor=1,
            hubdata=None
    ):
        self._hubdata = hubdata
        self._powermeter_factor = powermeter_factor
        self._warned_not_initialized = False
        self._is_initialized = False
        super().__init__(data_type, listenerentity, initval)

    @property
    def is_initialized(self) -> bool:
        if self._is_initialized is True:
            return True
        if isinstance(self.value, (float,int)) and self._hubdata.chargerobject.is_initialized:
            _LOGGER.debug(f"{CARPOWERSENSOR} has initialized")
            self._is_initialized = True
            return True
        return False

    @HubMember.value.setter
    def value(self, val):
        if val is None or val == 0:
            self._value = 0
        vval = try_parse(val, float)
        if not vval:
            vval = try_parse(val, int)
        if not vval:
            self._value = 0
        else:
            self._value = float(vval * self._powermeter_factor)





from dataclasses import dataclass, field
from ...models.chargertype.servicecalls_dto import ServiceCallsDTO
from ...models.chargertype.servicecalls_options import ServiceCallsOptions
from ...models.chargertype.calltype import CallType
from .const import (
    DOMAIN,
    ON,
    OFF,
    RESUME,
    PAUSE,
    PARAMS,
    UPDATECURRENT,
)


@dataclass
class ServiceCalls:
    domain:str
    model: ServiceCallsDTO
    options: ServiceCallsOptions
    on: CallType = field(init=False)
    off: CallType = field(init=False)
    pause: CallType = field(init=False)
    resume: CallType = field(init=False)
    update_current: CallType = field(init=False)

    def __post_init__(self):
        self.on = self.model.on
        self.off = self.model.off
        self.pause = self.model.pause
        self.resume = self.model.resume
        self.update_current = self.model.update_current

    def get_call(self, call) -> dict:
        ret = {DOMAIN: self.domain}
        calltype = self._get_call_type(call)
        ret[call] = calltype.call
        ret["params"] = calltype.params
        if call is UPDATECURRENT:
            if self.options.allowupdatecurrent is True:
                ret[PARAMS] = self.update_current.params
            else:
                raise AttributeError
        return ret

    def _get_call_type(self, call) -> CallType:
        _callsdict = {
            ON: self.on,
            OFF: self.off,
            PAUSE: self.pause,
            RESUME: self.resume,
            UPDATECURRENT: self.update_current
        }
        return _callsdict.get(call)



import logging
import time
from datetime import datetime
from ...models.chargerstates import CHARGERSTATES
from abc import abstractmethod
from typing import List, Dict

_LOGGER = logging.getLogger(__name__)


class ChargeControllerBase:  
    def __init__(
        self,
        charger_state_translation:Dict[CHARGERSTATES,List[str]],
        non_hours:List[int] = [],
        timeout:int = 300
    ):
        self._charger_state_translation = self._check_charger_states(charger_state_translation)
        self.done_timeout: int = timeout
        self._latest_charger_start: float = time.time()
        self._non_hours = non_hours

    @property
    def _is_timeout(self) -> bool:
        return time.time() - self.latest_charger_start > self._done_timeout

    @property
    def latest_charger_start(self) -> float:
        return self._latestchargerstart

    @latest_charger_start.setter
    def latest_charger_start(self, val):
        self._latestchargerstart = val

    @property
    @abstractmethod
    def status(self) -> str:
        if self._hub.is_initialized is False:
            return "Hub not ready. Check logs!"
        if self._hub.is_initialized is True:
            if self._chargecontroller_initalized is False:
                self._chargecontroller_initalized = True
                _LOGGER.debug("Chargecontroller is initialized and ready to work!")
        ret = self._get_status()
        if ret == CHARGERSTATES.Error:
            msg = f"Chargecontroller returned faulty state. Charger reported {self._hub.chargerobject.value.lower()} as state."
            _LOGGER.error(msg)
        return ret.name

    @property
    @abstractmethod
    def charger_state(self) -> CHARGERSTATES:
        """
        Check if you are below the starting threshold of charger
        :return: returns enum CHARGERSTATES
        """
        pass

    @property
    @abstractmethod
    def charger_done(self) -> bool:
        """
        
        :return: returns bool
        """
        pass

    @charger_done.setter
    @abstractmethod
    def charger_done(self, val) -> None:
        """
        
        :return: returns None
        """
        pass

    @property
    @abstractmethod
    def charger_enabled(self) -> bool:
        """
        
        :return: returns bool
        """
        pass

    @property
    @abstractmethod
    def free_charge(self) -> bool:
        """
        
        :return: returns bool
        """
        pass

    @abstractmethod
    def _get_status_charging(self) -> CHARGERSTATES:
        """
        
        :return: returns enum CHARGERSTATES
        """
        pass

    @abstractmethod
    def _get_status_connected(self) -> CHARGERSTATES:
        """
        
        :return: returns enum CHARGERSTATES
        """
        pass


    def _get_status(self):
        ret = CHARGERSTATES.Error
        update_timer = False
        charger_state = self._hub.chargerobject.value.lower()
        free_charge = self._hub.locale.data.free_charge(self._hub.locale.data)

        if self._hub.charger_enabled.value is False:
            update_timer = True
            ret = CHARGERSTATES.Disabled
        elif charger_state in self._hub.chargertype.charger.chargerstates[CHARGERSTATES.Done]:
            self._hub.charger_done.value = True
            ret = CHARGERSTATES.Done
        elif charger_state in self._hub.chargertype.charger.chargerstates[CHARGERSTATES.Idle]:
            update_timer = True
            ret = CHARGERSTATES.Idle
            if self._hub.charger_done.value is True:
                self._hub.charger_done.value = False
        elif charger_state not in self._hub.chargertype.charger.chargerstates[CHARGERSTATES.Idle] and self._hub.charger_done.value is True:
            ret = CHARGERSTATES.Done
        elif datetime.now().hour in self._hub.non_hours and free_charge is False and self._hub.timer.is_override is False:
            update_timer = True
            ret = CHARGERSTATES.Stop
        elif charger_state in self._hub.chargertype.charger.chargerstates[CHARGERSTATES.Connected]:
            ret = self._get_status_connected(charger_state)
            update_timer = (ret == CHARGERSTATES.Stop)
        elif charger_state in self._hub.chargertype.charger.chargerstates[CHARGERSTATES.Charging]:
            ret = self._get_status_charging()
            update_timer = True

        if update_timer is True:
            self.update_latestchargerstart()
        return ret

    def _is_done(self, charger_state) -> bool:
        if len(self._hub.chargertype.charger.chargerstates[CHARGERSTATES.Done]) > 0:
            return charger_state in self._hub.chargertype.charger.chargerstates[CHARGERSTATES.Done]
        return time.time() - self.latest_charger_start > self.done_timeout

    def update_latestchargerstart(self):
        self.latest_charger_start = time.time()

    def _below_start_threshold(
            self,
            predicted_energy: float,
            current_peak: float,
            threshold_start: float
    ) -> bool:
        """
        Check if you are below the starting threshold of charger
        :param predicted_energy: predicted energy for this hour in kWh
        :param current_peak: currently set energy peak, in kWh
        :param threshold_start: the allowed peak percentage to start. 0-1
        :return: returns bool
        """
        return (predicted_energy * 1000) < ((current_peak * 1000) * threshold_start)

    def _above_stop_threshold(
            self,
            predicted_energy: float,
            current_peak: float,
            threshold_stop: float
    ) -> bool:
        """
        Check if you are above the stopping threshold of charger
        :param predicted_energy: predicted energy for this hour in kWh
        :param current_peak: currently set energy peak, in kWh
        :param threshold_stop: the limit peak percentage to stop. 0-1
        :return: returns bool
        """
        return (predicted_energy * 1000) > ((current_peak * 1000) * threshold_stop)

    def _check_charger_states(self, input:Dict[CHARGERSTATES,List[str]]) -> Dict[CHARGERSTATES,List[str]]:
        if len(input) == 0:
            raise AssertionError
        return input


    

    



from .countries.default import Default
from .countries.belgium import VregBelgium
from .countries.sweden import SE_Bjerke_Energi, SE_FalbygdensEnergi, SE_Gothenburg, SE_Karlstad, SE_Kristinehamn, SE_Linde_Energi, SE_Malarenergi, SE_Malung_Salen, SE_Nacka_normal, SE_NACKA_timediff, SE_Partille, SE_SHE_AB, SE_Skovde, SE_Sollentuna
from .countries.norway import NO_AgderEnergi, NO_Elvia, NO_GlitreEnergi, NO_LNett, NO_Lede, NO_Mellom, NO_Tensio, NO_BKK


"""LOCALETYPES"""
LOCALE_SE_GOTHENBURG = "Gothenburg, Sweden"
LOCALE_SE_KARLSTAD = "Karlstad, Sweden"
LOCALE_SE_KRISTINEHAMN = "Kristinehamn, Sweden"
LOCALE_SE_NACKA_NORMAL = "Nacka, Sweden (Normal tariffe)"
#LOCALE_SE_NACKA_TIMEDIFF = "Nacka, Sweden (Time differentiated tariffe)"
LOCALE_SE_PARTILLE = "Partille, Sweden"
LOCALE_DEFAULT = "Other, just want to test"
LOCALE_SE_SALA = "Sala-Heby Energi AB, Sweden"
LOCALE_SE_MALUNG_SALEN = "Malung-Sälen, Sweden (Malungs elverk)"
LOCALE_SE_SKOVDE = "Skövde, Sweden"
LOCALE_SE_SOLLENTUNA = "Sollentuna Energi, Sweden"
LOCALE_BE_VREG = "Belgium (VREG)"
LOCALE_SE_BJERKE_ENERGI = "Bjärke Energi, Sweden"
LOCALE_NO_GLITRE_ENERGI = "Glitre Energi, Norway"
LOCALE_NO_AGDER_ENERGI = "Agder Energi, Norway"
LOCALE_NO_LNETT = "LNett, Norway"
LOCALE_NO_TENSIO = "Tensio, Norway"
LOCALE_SE_LINDE_ENERGI = "Linde Energi, Sweden"
LOCALE_SE_FALBYDGENS_ENERGI = "Falbygdens Energi, Sweden"
LOCALE_NO_ELVIA = "Elvia, Norway"
LOCALE_NO_LEDE = "Lede, Norway"
LOCALE_NO_BKK = "BKK, Norway"
LOCALE_SE_MALARENERGI = "Mälarenergi, Sweden"
LOCALE_NO_MELLOM = "Mellom, Norway"

LOCALETYPEDICT = {
    LOCALE_DEFAULT: Default,
    LOCALE_SE_GOTHENBURG: SE_Gothenburg,
    LOCALE_SE_PARTILLE: SE_Partille,
    LOCALE_SE_KARLSTAD: SE_Karlstad,
    LOCALE_SE_KRISTINEHAMN: SE_Kristinehamn,
    LOCALE_SE_NACKA_NORMAL: SE_Nacka_normal,
    LOCALE_SE_MALUNG_SALEN: SE_Malung_Salen,
    LOCALE_SE_SALA: SE_SHE_AB,
    LOCALE_SE_SKOVDE: SE_Skovde,
    LOCALE_SE_SOLLENTUNA: SE_Sollentuna,
    LOCALE_BE_VREG: VregBelgium,
    LOCALE_SE_BJERKE_ENERGI: SE_Bjerke_Energi,
    LOCALE_NO_GLITRE_ENERGI: NO_GlitreEnergi,
    LOCALE_NO_AGDER_ENERGI: NO_AgderEnergi,
    LOCALE_NO_LNETT: NO_LNett,
    LOCALE_NO_TENSIO: NO_Tensio,
    LOCALE_SE_LINDE_ENERGI: SE_Linde_Energi,
    LOCALE_SE_FALBYDGENS_ENERGI: SE_FalbygdensEnergi,
    LOCALE_NO_ELVIA: NO_Elvia,
    LOCALE_NO_LEDE: NO_Lede,
    LOCALE_NO_BKK: NO_BKK,
    LOCALE_SE_MALARENERGI: SE_Malarenergi,
    LOCALE_NO_MELLOM: NO_Mellom
}

"""Lookup locales for config flow"""
LOCALES = [
    LOCALE_BE_VREG,
    LOCALE_NO_AGDER_ENERGI,
    LOCALE_NO_BKK,
    LOCALE_NO_ELVIA,
    LOCALE_NO_GLITRE_ENERGI,
    LOCALE_NO_LEDE,
    LOCALE_NO_LNETT,
    LOCALE_NO_MELLOM,
    LOCALE_NO_TENSIO,
    LOCALE_SE_BJERKE_ENERGI,
    LOCALE_SE_FALBYDGENS_ENERGI,
    LOCALE_SE_GOTHENBURG,
    LOCALE_SE_KRISTINEHAMN,
    LOCALE_SE_LINDE_ENERGI,
    LOCALE_SE_MALUNG_SALEN,
    LOCALE_SE_MALARENERGI,
    LOCALE_SE_NACKA_NORMAL,
    LOCALE_SE_PARTILLE,
    LOCALE_SE_SALA,
    LOCALE_SE_SKOVDE,
    LOCALE_SE_SOLLENTUNA,
    LOCALE_DEFAULT
    ]


class LocaleData:
    def __init__(self, input_type:str, domain:str):
        self._data = None
        self._type = input_type
        self._domain = domain

        self._data = LOCALETYPEDICT[input_type]

    @property
    def type(self) -> str:
        return self._type

    @property
    def data(self):
        return self._data



from ...models.chargerstates import CHARGERSTATES
from .chargecontrollerbase import ChargeControllerBase
#from custom_components.peaqev.peaqservice.chargecontroller.chargecontrollerbase import ChargeControllerBase


class ChargeController(ChargeControllerBase):
    def __init__(self, hub):
        #super().__init__(hub)
        super().__init__(charger_state_translation=self.self._hub.chargertype.charger.chargerstates)
        #self._core = _core(charger_state_translation=self._hub.chargertype.charger.chargerstates)

    @property
    def below_startthreshold(self) -> bool:
        return self._below_start_threshold(
            predicted_energy=self._hub.prediction.predictedenergy,
            current_peak=self._hub.current_peak_dynamic,
            threshold_start=self._hub.threshold.start/100
        )

    @property
    def above_stopthreshold(self) -> bool:
        return self._above_stop_threshold(
            predicted_energy=self._hub.prediction.predictedenergy,
            current_peak=self._hub.current_peak_dynamic,
            threshold_stop=self._hub.threshold.stop/100
        )

    def _get_status_charging(self) -> CHARGERSTATES:
        if self.above_stopthreshold and self._hub.totalhourlyenergy.value > 0 and self._hub.locale.data.free_charge(self._hub.locale.data) is False:
            ret = CHARGERSTATES.Stop
        else:
            ret = CHARGERSTATES.Start
        return ret

    def _get_status_connected(self, charger_state) -> CHARGERSTATES:
        if self._hub.carpowersensor.value < 1 and self._is_done(charger_state):
            ret = CHARGERSTATES.Done
        else:
            if (self.below_startthreshold and self._hub.totalhourlyenergy.value != 0) or self._hub.locale.data.free_charge(self._hub.locale.data) is True:
                ret = CHARGERSTATES.Start
            else:
                ret = CHARGERSTATES.Stop
        return ret



from .hoursbase import Hours

class RegularHours(Hours):
    def __init__(self, hub):
        self.hub = hub
        super().__init__(False, self.hub.options.nonhours, self.hub.options.cautionhours)

    @property
    def nordpool_entity(self):
        pass

    def update_nordpool(self):
        pass

    @property
    def dynamic_caution_hours(self) -> dict:
        pass

    @property
    def is_initialized(self) -> bool:
        return True

    @property
    def options(self):
        pass


from ..services.hourselection.hourselectionfactory import HourselectionFactory
from ..services.timer.timer import Timer
from ..services.chargecontroller.chargecontrollerfactory import ChargeControllerFactory
from ..services.threshold.thresholdfactory import ThresholdFactory
from ..services.chargecontroller.chargecontrollerbase import ChargeControllerBase
from ..services.hourselection.hoursbase import Hours
from ..services.prediction.prediction import Prediction
from ..services.production.production import ProductionService
from ..services.scheduler.scheduler import Scheduler
from ..services.session.session import SessionPrice
from ..services.threshold.threshold import ThresholdBase
from ..services.locale.Locale import LocaleData
from .hub_sensors import HubSensorsFactory
from .hub_options import HubOptions

class Hub:
    def __init__(self, options: HubOptions, domain: str, state_machine, chargerobj):
        self.state_machine = state_machine
        self.domain = domain

        self.options: HubOptions = options
        self.sensors = HubSensorsFactory.create(self.options)
        self.timer: Timer = Timer()
        self.hours: Hours = HourselectionFactory.create(self)
        self.threshold = ThresholdFactory.create(self)
        self.prediction = Prediction(self)

        #chargecontroller: ChargeControllerBase = field(init=False)

        self.sensors.setup(state_machine=state_machine, options=options, domain=domain, chargerobject=chargerobj)
        self.sensors.init_hub_values()





from datetime import datetime
import statistics as stat
from .top_up import top_up, TopUpDTO
from ...models.hourselection.const import (
    CAUTIONHOURTYPE_SUAVE,
    CAUTIONHOURTYPE_INTERMEDIATE,
    CAUTIONHOURTYPE
)
from enum import Enum
from .hoursselection_helpers import HourSelectionHelpers
from ...models.hourselection.hourobject import HourObject, HourObjectExtended
from ...models.hourselection.hourselectionmodels import HourSelectionModel, HourSelectionOptions

class HourTypeList(Enum):
    NonHour = 0
    CautionHour = 1
    DynCautionHour = 2


class Hoursselection:
    def __init__(
            self,      
            absolute_top_price: float = 0,
            min_price: float = 0,
            cautionhour_type: float = CAUTIONHOURTYPE[CAUTIONHOURTYPE_SUAVE],
            allow_top_up: bool = False,
            base_mock_hour: int = None
    ):
        self.model = HourSelectionModel(
            options=HourSelectionOptions(
                cautionhour_type=cautionhour_type, 
                absolute_top_price=self._set_absolute_top_price(absolute_top_price), 
                min_price=min_price, 
                allow_top_up=allow_top_up
                )
            )
        self._base_mock_hour: int = base_mock_hour
        self._validate()
    
    def _set_absolute_top_price(self, val) -> float:
        if val is None:
            return float("inf")
        if val <= 0:
            return float("inf")
        return float(val)

    def _validate(self):
        assert 0 < self.model.options.cautionhour_type <= 1

    @property
    def options(self) -> HourSelectionOptions:
        return self.model.options

    @property
    def non_hours(self) -> list:
        self._update_hour_lists(self.model.hours.hours_today, self.model.hours.hours_tomorrow, listtype=HourTypeList.NonHour)
        return self.model.hours.non_hours

    @property
    def caution_hours(self) -> list:
        self._update_hour_lists(self.model.hours.hours_today, self.model.hours.hours_tomorrow, listtype=HourTypeList.CautionHour)
        return self.model.hours.caution_hours

    @property
    def dynamic_caution_hours(self) -> dict:
        self._update_hour_lists(self.model.hours.hours_today, self.model.hours.hours_tomorrow, listtype=HourTypeList.DynCautionHour)
        return self.model.hours.dynamic_caution_hours

    @property
    def prices(self) -> list:
        return self.model.prices_today

    @prices.setter
    def prices(self, val):
        self.model.prices_today = val
        if self.prices == self.prices_tomorrow:
            self.prices_tomorrow = None
        else:
            self.update()

    @property
    def prices_tomorrow(self) -> list:
        return self.model.prices_tomorrow

    @prices_tomorrow.setter
    def prices_tomorrow(self, val):
        self.model.prices_tomorrow = HourSelectionHelpers._convert_none_list(val)
        self.update()

    def update(self, testhour:int = None):
        today_ready = self._update_per_day(self.prices)
        hours_today = self._add_remove_limited_hours(today_ready)
        hours_tomorrow = HourObject([],[],dict())
        
        if self.prices_tomorrow is not None and len(self.prices_tomorrow) > 0:
            hours_tomorrow = self._add_remove_limited_hours(
                self._update_per_day(self.prices_tomorrow)
                )

        self.model.hours.hours_today = hours_today
        self.model.hours.hours_tomorrow = hours_tomorrow
        self._update_hour_lists(hours_today, hours_tomorrow, testhour)
        
    def _set_top_up(self, testhour: int = None) -> None:
        if self.model.options.allow_top_up is True and self.prices_tomorrow is not None and len(self.prices_tomorrow) > 0:
            ret = top_up(TopUpDTO(
                self.model.hours.non_hours, 
                self.model.hours.caution_hours, 
                self.model.hours.dynamic_caution_hours, 
                self.model.options.absolute_top_price,
                self.model.options.min_price,
                testhour if testhour is not None else self._base_mock_hour if self._base_mock_hour is not None else datetime.now().hour, 
                self.model.hours.hours_today, 
                self.model.hours.hours_tomorrow, 
                self.model.prices_today, 
                self.model.prices_tomorrow
                ))
            self.model.hours.non_hours = ret.nh
            self.model.hours.caution_hours = ret.ch
            self.model.hours.dynamic_caution_hours = ret.dyn_ch

    def _update_hour_lists(self, hours_today:HourObject, hours_tomorrow:HourObject, testhour:int = None, listtype:HourTypeList = None) -> None:
        hour = testhour if testhour is not None else self._base_mock_hour if self._base_mock_hour is not None else datetime.now().hour

        if listtype is not None:
            match listtype:
                case HourTypeList.NonHour:
                    self._update_nonhour_list(hours_today, hours_tomorrow, hour)
                    print(self.model.hours.non_hours)
                case HourTypeList.CautionHour:
                    self._update_cautionhour_list(hours_today, hours_tomorrow, hour)
                case HourTypeList.DynCautionHour:
                    self._update_dyn_cautionhour_dict(hours_today, hours_tomorrow, hour)
                case _:
                    pass
            self._set_top_up(hour)
        else:
            self._update_nonhour_list(hours_today, hours_tomorrow, hour)
            self._update_cautionhour_list(hours_today, hours_tomorrow, hour)
            self._update_dyn_cautionhour_dict(hours_today, hours_tomorrow, hour)
        
    def _update_nonhour_list(self, hours_today:HourObject, hours_tomorrow:HourObject, hour:int) -> None:
        self.model.hours.non_hours = []
        self.model.hours.non_hours.extend(h for h in hours_today.nh if h >= hour)
        self.model.hours.non_hours.extend(h for h in hours_tomorrow.nh if h < hour)

    def _update_cautionhour_list(self, hours_today:HourObject, hours_tomorrow:HourObject, hour:int) -> None:
        self.model.hours.caution_hours = []
        self.model.hours.caution_hours.extend(h for h in hours_today.ch if h >= hour)
        self.model.hours.caution_hours.extend(h for h in hours_tomorrow.ch if h < hour)

    def _update_dyn_cautionhour_dict(self, hours_today:HourObject, hours_tomorrow:HourObject, hour:int) -> None:
        self.model.hours.dynamic_caution_hours = dict()
        self.model.hours.dynamic_caution_hours.update({k: v for k, v in hours_today.dyn_ch.items() if k >= hour})
        self.model.hours.dynamic_caution_hours.update({k: v for k, v in hours_tomorrow.dyn_ch.items() if k < hour})

    def _update_per_day(self, prices) -> HourObjectExtended:
        pricedict = dict
        if prices is not None and len(prices) > 1:
            pricedict = HourSelectionHelpers._create_dict(prices)
            normalized_pricedict = HourSelectionHelpers._create_dict(
                HourSelectionHelpers._normalize_prices(prices)
                )
            if stat.stdev(prices) > 0.05:
                prices_ranked = HourSelectionHelpers._rank_prices(pricedict, normalized_pricedict)
                ready_hours = self._determine_hours(prices_ranked, prices)
                return HourObjectExtended(
                    ready_hours.nh, 
                    ready_hours.ch, 
                    ready_hours.dyn_ch, 
                    pricedict
                    )
            return HourObjectExtended([], [], dict(), pricedict)
        
    def _add_remove_limited_hours(self, hours: HourObjectExtended) -> HourObject:
        """Removes cheap hours and adds expensive hours set by user limitation"""
        if hours is None:
            return HourObject([],[],dict())

        if self.model.options.absolute_top_price is not None:
                ret = self._add_expensive_non_hours(hours)
        else: 
            ret = HourObject(hours.nh, hours.ch, hours.dyn_ch)
        if self.model.options.min_price > 0:
            ret = self._remove_cheap_hours(hours)
        return ret

    def _remove_cheap_hours(self, hours: HourObjectExtended) -> HourObject:
        lst = (h for h in hours.pricedict if hours.pricedict[h] < self.model.options.min_price)
        for h in lst:
            if h in hours.nh:
                hours.nh.remove(h)
            elif h in hours.ch:
                hours.ch.remove(h)
                hours.dyn_ch.pop(h)    
        return HourObject(hours.nh, hours.ch, hours.dyn_ch)

    def _add_expensive_non_hours(self, readyhours:HourObjectExtended) -> HourObject:
        lst = (h for h in readyhours.pricedict if readyhours.pricedict[h] >= self.model.options.absolute_top_price)
        for h in lst:
            if h not in readyhours.nh:
                readyhours.nh.append(h)
                if h in readyhours.ch:
                    readyhours.ch.remove(h)
                if len(readyhours.dyn_ch) > 0:
                    if h in readyhours.dyn_ch.keys():
                        readyhours.dyn_ch.pop(h)
        readyhours.nh.sort()
        return HourObject(readyhours.nh, readyhours.ch, readyhours.dyn_ch)

    def _determine_hours(self, price_list: dict, prices: list) -> HourObject:
        _nh = []
        _dyn_ch = dict()
        _ch = []
        for p in price_list:
            _permax = round(abs(price_list[p]["permax"] - 1), 2)
            if self.model.options.cautionhour_type == CAUTIONHOURTYPE[CAUTIONHOURTYPE_SUAVE]:
                _permax += 0.15
            elif self.model.options.cautionhour_type == CAUTIONHOURTYPE[CAUTIONHOURTYPE_INTERMEDIATE]:
                _permax += 0.05

            if float(price_list[p]["permax"]) <= self.model.options.cautionhour_type or float(price_list[p]["val"]) <= (sum(prices)/len(prices)):
                _ch.append(p)
                _dyn_ch[p] = round(_permax,2)
            else:
                _nh.append(p)

        return HourObject(_nh, _ch, _dyn_ch)
    
    def get_average_kwh_price(self, testhour:int = None):
        ret = self._get_charge_or_price(testhour=testhour)
        return round(sum(ret.values())/len(ret),2)
        
    def get_total_charge(self, currentpeak:float, testhour:int = None) -> float:
        ret = self._get_charge_or_price(currentpeak, testhour)
        return round(sum(ret.values()),1)

    def _get_charge_or_price(self, currentpeak:float = None, testhour:int = None) -> dict:
        hour = self._set_hour(testhour)
        ret = dict()

        def _looper_charge(h:int):
            if h in self.model.hours.dynamic_caution_hours:
                    ret[h] = self.model.hours.dynamic_caution_hours[h] * currentpeak
            elif h in self.model.hours.non_hours:
                ret[h] = 0
            else:
                ret[h] = currentpeak

        def _looper_price(h:int, tomorrow_active:bool):
            if h in self.model.hours.dynamic_caution_hours:
                    if tomorrow_active:
                        if h < hour and len(self.prices_tomorrow) > 0:
                            ret[h] = self.model.hours.dynamic_caution_hours[h] * self.prices_tomorrow[h]
                    if h >= hour:
                        ret[h] = self.model.hours.dynamic_caution_hours[h] * self.prices[h]
            elif h not in self.model.hours.non_hours:
                if h < hour and len(self.prices_tomorrow) > 0:
                    ret[h] = self.prices_tomorrow[h]
                if h >= hour:
                    ret[h] = self.prices[h]

        if self.prices_tomorrow is None:
            for h in range(hour,24):
                if currentpeak is not None:
                    _looper_charge(h)
                else:
                    _looper_price(h, False)
        else:
            for h in range(hour,(hour+24)):
                h = h-24 if h > 23 else h
                if currentpeak is not None:
                    _looper_charge(h)
                else:
                    _looper_price(h, True)
        return ret

    def _set_hour(self, testhour:int = None) -> int:
        return datetime.now().hour if testhour is None else testhour

